import io
import smtplib
import typing
import zipfile
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path


__all__ = ["SimpleEmail"]


class SimpleEmail:
    """SimpleEmail Sender.

    Demo::

        with SimpleEmail("smtp.gmail.com", 465, "someone@gmail.com", "PASSWORD") as s:
        print(
            s.send_message(
                "This is Title",
                "This is body text or file path(.md/.txt)",
                "Author<someone@gmail.com>",
                "anybody@gmail.com",
                files="a.py,b.py,c.txt",
                filename="files.zip",
                encoding="u8",
            )
        )
    """

    def __init__(self, host, port, user, pwd, smtp_cls=smtplib.SMTP_SSL):
        self.host = host
        self.port = port
        self.user = user
        self.pwd = pwd
        self.smtp_cls = smtp_cls

        self.server = None

    def __enter__(self):
        self.server = self.smtp_cls(self.host, self.port)
        self.server.__enter__()
        self.server.login(self.user, self.pwd)
        return self

    def __exit__(self, *_):
        if self.server:
            return self.server.__exit__(*_)

    def _send_message(self, msg):
        if not self.server:
            raise RuntimeError("with context not entered")
        return self.server.send_message(msg)

    def prepare_attach(self, files: str, filename="files.zip"):
        with io.BytesIO() as zip_buffer:
            with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zipf:
                total_size = 0
                for _path in files.split(","):
                    file_path = Path(_path)
                    if file_path.is_file():
                        zipf.write(file_path.as_posix(), arcname=file_path.name)
                        total_size += file_path.stat().st_size
                    else:
                        raise FileNotFoundError(file_path.as_posix())
                if not total_size:
                    return None
            zip_buffer.seek(0)
            part = MIMEBase("application", "octet-stream")
            content = zip_buffer.read()
            part.set_payload(content)
            encoders.encode_base64(part)
            part.add_header("Content-Disposition", "attachment", filename=filename)
            return part

    def send_message(
        self,
        title: str,
        body: typing.Union[str, Path],
        from_email: str,
        to_emails: str,
        files="",
        filename: typing.Union[str, tuple] = "files.zip",
        encoding=None,
    ):
        to_mail_list = [i.strip() for i in to_emails.split(",")]
        if not to_mail_list or not to_mail_list[0]:
            raise ValueError("email is %s" % to_emails)
        msg = MIMEMultipart()
        msg["From"] = from_email
        msg["To"] = ", ".join(to_mail_list)
        msg["Subject"] = title
        try:
            # get text from path
            b_path = Path(body)
            if b_path.is_file() and b_path.suffix in (".txt", ".md"):
                body = b_path.read_text(encoding=encoding)
        except (UnicodeDecodeError, OSError):
            pass
        if files:
            attach_part = self.prepare_attach(files, filename=filename)
        else:
            attach_part = None
        msg.attach(MIMEText(str(body), "plain"))
        if attach_part:
            msg.attach(attach_part)
        return self._send_message(msg)



# -*- coding: utf-8 -*-
from .main import create_app, __version__, ZipApp, pip_install_target
from .activate_zipapps import activate

__all__ = ["create_app", "activate", "__version__", "ZipApp", "pip_install_target"]
__doc__ = "Package your python code into one zip file, even a virtual environment."



import argparse
import json
import sys
import time
import typing
from pathlib import Path

from . import __version__
from .main import ZipApp

USAGE = r"""
===========================================================================
0. package your code without any requirements

> python3 -m zipapps -c -a ./simple_package -p /usr/bin/python3 -o simple_package.pyz
> ./simple_package.pyz
OR
> python3 -m zipapps -c -a ./simple_package -o simple_package.pyz
> python3 simple_package.pyz

------------------------------------

1. use zipapps to make a standalone app which need bottle installed
> python3 -m zipapps -c -a bottle_server.py -m bottle_server:main bottle
> python3 app.pyz
OR
> python3 -m zipapps -c -a bottle_server.py -m bottle_server:main bottle
> python3 app.pyz

------------------------------------

2. use zipapps for venv
> python3 -m zipapps -c bottle
> python3 app.pyz bottle_server.py
OR
> python3 -m zipapps -c -p /usr/bin/python3 bottle
> ./app.pyz bottle_server.py

------------------------------------

3. advanced usages
3.1 more args
> python3 -m zipapps -c -a package1,package2 -o server.pyz -m package1.server:main -p /usr/bin/python3 -r requirements.txt
> ./server.pyz

3.2 unzip C-libs to cache folder for zipimport do not support .so .pyd files.
    bad
        > python3 -m zipapps -c lxml
        > python3 app.pyz -c "import lxml.html;print(lxml.html.__file__)"
    good
        > python3 -m zipapps -c -u lxml lxml
        > python3 app.pyz -c "import lxml.html;print(lxml.html.__file__)"

PS: all the unknown args will be used by "pip install".
==========================================================================="""

PIP_PYZ_URL = "https://bootstrap.pypa.io/pip/pip.pyz"
DOWNLOAD_PYTHON_URL = "https://www.github.com/indygreg/python-build-standalone"


def _get_now():
    return time.strftime("%Y-%m-%d %H:%M:%S")


def _get_pth_path():
    py_exe_path = Path(sys.executable)
    for _path in py_exe_path.parent.glob("*._pth"):
        _pth_path = _path
        break
    else:
        fname = f"python{sys.version_info.major}{sys.version_info.minor}._pth"
        _pth_path = py_exe_path.parent / fname
    return _pth_path


def _append_pth():
    import re

    _pth_path = _get_pth_path()
    if _pth_path.is_file():
        print("find _pth file:", _pth_path.as_posix(), flush=True, file=sys.stderr)
        _path_bytes = _pth_path.read_bytes()
    else:
        _path_bytes = b""
    if not re.search(b"^import site$", _path_bytes):
        _path_bytes += b"\nimport site\n"
    if not re.search(b"^pip\.pyz$", _path_bytes):
        _path_bytes += b"\npip.pyz\n"
    _pth_path.write_bytes(_path_bytes)


def download_pip_pyz(target: typing.Optional[Path] = None, log=True):
    from urllib.request import urlretrieve

    pip_pyz_path = Path(target or (Path(sys.executable).parent / "pip.pyz"))
    if log:
        msg = f"Download {PIP_PYZ_URL} -> {pip_pyz_path.absolute().as_posix()}"
        print(_get_now(), msg, flush=True, file=sys.stderr)
    if pip_pyz_path.is_file():
        pip_pyz_path.unlink()
    urlretrieve(
        url=PIP_PYZ_URL,
        filename=pip_pyz_path.absolute().as_posix(),
        reporthook=lambda a, b, c: print(
            _get_now(),
            f"Downloading {int(100*(1+a)*b/c)}%, {(1 + a) * b} / {c}",
            end="\r",
            flush=True,
            file=sys.stderr,
        )
        if log
        else None,
    )
    try:
        sys.path.append(pip_pyz_path.absolute().as_posix())
        import pip

        assert pip

        if log:
            print(f"\n{_get_now()} install pip ok", flush=True, file=sys.stderr)
        return True
    except ImportError:
        if log:
            print(f"\n{_get_now()} install pip failed", flush=True, file=sys.stderr)


def handle_win32_embeded():
    _pth_path = _get_pth_path()

    if not _pth_path.is_file():
        return
    try:
        import pip

        assert pip

        return
    except ImportError:
        need_install = (
            input(f'\n{"=" * 50}\npip module not found, try installing?(Y/n)')
            .lower()
            .strip()
            or "y"
        )
        if need_install != "y":
            return
    print("find _pth file:", _pth_path.as_posix(), flush=True, file=sys.stderr)
    if download_pip_pyz():
        _append_pth()
    import os

    os.system("%s -m pip -V" % Path(sys.executable).as_posix())
    os.system("pause")


def main():
    parser = argparse.ArgumentParser(usage=USAGE, prog="Zipapps")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--output",
        "-o",
        default=ZipApp.DEFAULT_OUTPUT_PATH,
        help="The path of the output file, defaults to"
        f' "{ZipApp.DEFAULT_OUTPUT_PATH}".',
    )
    parser.add_argument(
        "--python",
        "-p",
        dest="interpreter",
        default=None,
        help="The path of the Python interpreter which will be "
        "set as the `shebang line`, defaults to `None`. With shebang `/usr/bin/python3` you can run app with `./app.pyz` directly, no need for `python3 app.pyz`",
    )
    parser.add_argument(
        "--main",
        "-m",
        default="",
        help="The entry point function of the application, "
        "the format is: `package` | `package.module` | `package.module:function` | `module:function`",
    )
    parser.add_argument(
        "--compress",
        "-c",
        dest="compressed",
        action="store_true",
        help="compress files with the deflate method or not.",
    )
    parser.add_argument(
        "--includes",
        "--add",
        "-a",
        default="",
        help="The given paths will be copied to `cache_path` while packaging, "
        'which can be used while running. The path strings will be splited by ",". '
        "such as `my_package_dir,my_module.py,my_config.json`, often used for libs not from `pypi` or some special config files",
    )
    parser.add_argument(
        "--unzip",
        "-u",
        default="",
        help='The names which need to be unzipped while running, splited by "," '
        '`without ext`, such as `bottle,aiohttp`, or the complete path like `bin/bottle.py,temp.py`. For `.so/.pyd` files(which can not be loaded by zipimport), or packages with operations of static files. if unzip is set to "*", then will unzip all files and folders. if unzip is set to **AUTO**, then will add the `.pyd` and `.so` files automatically. Can be overwrite with environment variable `ZIPAPPS_UNZIP`',
    )
    parser.add_argument(
        "--unzip-exclude",
        "-ue",
        default="",
        dest="unzip_exclude",
        help="The opposite of `--unzip` / `-u` which will not be unzipped, "
        "should be used with `--unzip` / `-u`. Can be overwrite with environment variable `ZIPAPPS_UNZIP_EXCLUDE`",
    )
    parser.add_argument(
        "--unzip-path",
        "-up",
        default="",
        help="If `unzip` arg is not null, cache files will be unzipped to the "
        "given path while running. Defaults to `zipapps_cache`, support some internal variables: `$TEMP/$HOME/$SELF/$PID/$CWD` as internal variables, for example `$HOME/zipapps_cache`. `$TEMP` means `tempfile.gettempdir()`, `$HOME` means `Path.home()`, `$SELF` means `.pyz` file path, `$PID` means `os.getpid()`, `$CWD` means `Path.cwd()`.",
    )
    parser.add_argument(
        "-cc",
        "--pyc",
        "--compiled",
        action="store_true",
        dest="compiled",
        help="Compile .py to .pyc for fast import, but zipapp does not work "
        "unless you unzip it.",
    )
    parser.add_argument(
        "--cache-path",
        "--source-dir",
        "-cp",
        default=None,
        dest="cache_path",
        help="The cache path of zipapps to store "
        "site-packages and `includes` files, "
        "which will be treat as PYTHONPATH. If not set, will "
        "create and clean-up in TEMP dir automately.",
    )
    parser.add_argument(
        "--shell",
        "-s",
        action="store_true",
        help="Only while `main` is not set, used for shell=True"
        " in subprocess.Popen.",
    )
    parser.add_argument(
        "--main-shell",
        "-ss",
        action="store_true",
        dest="main_shell",
        help="Only for `main` is not null, call `main` with subprocess.Popen: "
        '`python -c "import a.b;a.b.c()"`. This is used for `psutil` ImportError of DLL load.',
    )
    parser.add_argument(
        "--strict-python-path",
        "-spp",
        action="store_true",
        dest="ignore_system_python_path",
        help="Ignore global PYTHONPATH, only use zipapps_cache and app.pyz.",
    )
    parser.add_argument(
        "-b",
        "--build-id",
        default="",
        dest="build_id",
        help="a string to skip duplicate builds,"
        ' it can be the paths of files/folders which splited by ",", then the modify time will be used as build_id. If build_id contains `*`, will use `glob` function to get paths. For example, you can set requirements.txt as your build_id by `python3 -m zipapps -b requirements.txt -r requirements.txt` when you use pyz as venv.',
    )
    parser.add_argument(
        "--zipapps",
        "--env-paths",
        default="",
        dest="env_paths",
        help="Default --zipapps arg if it is not given while running."
        " Also support $TEMP/$HOME/$SELF/$PID/$CWD prefix, separated by commas.",
    )
    parser.add_argument(
        "--delay",
        "-d",
        "--lazy-pip",
        "--lazy-install",
        "--lazy-pip-install",
        action="store_true",
        dest="lazy_install",
        help="Install packages with pip while running, which means "
        "requirements will not be install into pyz file. Default unzip path will be changed to `SELF/zipapps_cache`",
    )
    parser.add_argument(
        "-pva",
        "--python-version-accuracy",
        "--python-version-slice",
        default=2,
        type=int,
        dest="python_version_slice",
        help="Only work for lazy-install mode, then `pip` target folders differ "
        "according to sys.version_info[:_slice], defaults to 2, which means "
        "3.8.3 equals to 3.8.4 for same version accuracy 3.8",
    )
    parser.add_argument(
        "--sys-paths",
        "--sys-path",
        "--py-path",
        "--python-path",
        default="",
        dest="sys_paths",
        help="Paths be insert to sys.path[0] while running."
        " Support $TEMP/$HOME/$SELF/$PID/$CWD prefix, separated by commas.",
    )
    parser.add_argument(
        "--activate",
        default="",
        dest="activate",
        help="Activate the given paths of zipapps app, "
        "only activate them but not run them, separated by commas.",
    )
    parser.add_argument(
        "--ensure-pip",
        action="store_true",
        dest="ensure_pip",
        help="Add the ensurepip package to your pyz file, works for "
        "embed-python(windows) or other python versions without `pip`"
        " installed but `lazy-install` mode is enabled. [EXPERIMENTAL]",
    )
    parser.add_argument(
        "--layer-mode",
        action="store_true",
        dest="layer_mode",
        help="Layer mode for the serverless use case, "
        "__main__.py / ensure_zipapps.py / activate_zipapps.py files will not be set in this mode.",
    )
    parser.add_argument(
        "--layer-mode-prefix",
        default="python",
        dest="layer_mode_prefix",
        help="Only work while --layer-mode is set, "
        "will move the files in the given prefix folder.",
    )
    parser.add_argument(
        "-czc",
        "--clear-zipapps-cache",
        action="store_true",
        dest="clear_zipapps_cache",
        help="Clear the zipapps cache folder after running, "
        "but maybe failed for .pyd/.so files.",
    )
    parser.add_argument(
        "-czs",
        "--clear-zipapps-self",
        action="store_true",
        dest="clear_zipapps_self",
        help="Clear the zipapps pyz file self after running.",
    )
    parser.add_argument(
        "--chmod",
        default="",
        dest="chmod",
        help="os.chmod(int(chmod, 8)) for unzip files with `--chmod=777`,"
        " unix-like system only",
    )
    parser.add_argument(
        "--dump-config",
        default="",
        dest="dump_config",
        help="Dump zipapps build args into JSON string."
        " A file path needed and `-` means stdout.",
    )
    parser.add_argument(
        "--load-config",
        default="",
        dest="load_config",
        help="Load zipapps build args from a JSON file.",
    )
    parser.add_argument(
        "--freeze-reqs",
        default="",
        dest="freeze",
        help="Freeze package versions of pip args with venv,"
        " output to the given file path.",
    )
    parser.add_argument(
        "-q",
        "--quite",
        action="count",
        dest="quite_mode",
        help="mute logs.",
    )
    parser.add_argument(
        "--download-pip-pyz",
        default="",
        dest="download_pip_pyz",
        help=f'Download pip.pyz from "{PIP_PYZ_URL}"',
    )
    parser.add_argument(
        "--download-python",
        action="store_true",
        dest="download_python",
        help=f'(please use `python -m zipapps.download_python` instead). Download standalone python from "{DOWNLOAD_PYTHON_URL}"',
    )
    parser.add_argument(
        "--rm-patterns",
        default="*.dist-info,__pycache__",
        dest="rm_patterns",
        help='Delete useless files or folders, splited by "," and defaults to `*.dist-info,__pycache__`. Recursively glob: **/*.pyc',
    )
    if len(sys.argv) == 1:
        parser.print_help()
        handle_win32_embeded()
        return
    args, pip_args = parser.parse_known_args()
    if args.download_python:
        raise ValueError("please use `python -m zipapps.download_python` instead")
    elif args.download_pip_pyz:
        return download_pip_pyz(args.download_pip_pyz)
    if args.quite_mode:
        ZipApp.LOGGING = False
        if "-q" not in pip_args and "--quiet" not in pip_args:
            pip_args.append(f'-{"q" * args.quite_mode}')
    ZipApp._log(f"zipapps args: {args}, pip install args: {pip_args}")
    if args.activate:
        from .activate_zipapps import activate

        for path in args.activate.split(","):
            activate(path)
        return
    if args.freeze:
        from .freezing import FreezeTool

        with FreezeTool(args.freeze, pip_args) as ft:
            ft.run()
        return
    if args.load_config:
        with open(args.load_config, "r", encoding="utf-8") as f:
            kwargs = json.load(f)
            app = ZipApp(**kwargs)
    else:
        app = ZipApp(
            includes=args.includes,
            cache_path=args.cache_path,
            main=args.main,
            output=args.output,
            interpreter=args.interpreter,
            compressed=args.compressed,
            shell=args.shell,
            unzip=args.unzip,
            unzip_path=args.unzip_path,
            ignore_system_python_path=args.ignore_system_python_path,
            main_shell=args.main_shell,
            pip_args=pip_args,
            compiled=args.compiled,
            build_id=args.build_id,
            env_paths=args.env_paths,
            lazy_install=args.lazy_install,
            sys_paths=args.sys_paths,
            python_version_slice=int(args.python_version_slice),
            ensure_pip=args.ensure_pip,
            layer_mode=args.layer_mode,
            layer_mode_prefix=args.layer_mode_prefix,
            clear_zipapps_cache=args.clear_zipapps_cache,
            unzip_exclude=args.unzip_exclude,
            chmod=args.chmod,
            clear_zipapps_self=args.clear_zipapps_self,
            rm_patterns=args.rm_patterns,
        )
    if args.dump_config:
        config_json = json.dumps(app.kwargs)
        if args.dump_config == "-":
            print(config_json)
        else:
            with open(args.dump_config, "w", encoding="utf-8") as f:
                f.write(config_json)
    else:
        return app.build()


if __name__ == "__main__":
    main()



import argparse
import http.client
import json
import re
import shutil
import tarfile
import time
import traceback
import urllib.request
from collections import deque
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Asset(object):
    # cpython-3.9.19+20240726-x86_64_v4-unknown-linux-musl-lto-full.tar.zst
    name: str
    # keywords=['cpython', '3.9.19', 'x86_64_v4', 'unknown', 'linux', 'musl', 'noopt-full.tar.zst']
    keywords: str
    # https://github.com/indygreg/python-build-standalone/releases/download/20240726/cpython-3.9.19%2B20240726-x86_64_v4-unknown-linux-musl-noopt-full.tar.zst
    url: str
    # 27204188
    size: int


def get_assets():
    api = (
        "https://api.github.com/repos/indygreg/python-build-standalone/releases/latest"
    )
    req = urllib.request.Request(api, headers={"User-Agent": "Chrome"})
    with urllib.request.urlopen(url=req, timeout=10) as resp:
        data = json.loads(resp.read())
        # print(data)
        urls = [
            Asset(
                i["name"],
                [k.replace(f"+{data['name']}", "") for k in i["name"].split("-", 6)],
                i["browser_download_url"],
                i["size"],
            )
            for i in data["assets"]
            if not i["name"].endswith(".sha256") and i["name"].count("-") >= 6
        ]
        return urls


def get_time():
    return time.strftime("%H:%M:%S")


def choose_asset(assets: list, keywords, auto=False):
    keywords = {k for k in keywords if k}
    not_include = set()
    if auto:
        import platform

        system = platform.system().lower()
        keywords.add(system)
        if system == "windows":
            keywords.add("msvc-shared")
        elif system == "linux":
            keywords.add("gnu")
            keywords.add("unknown")
            not_include.add("v2")
            not_include.add("v3")
            not_include.add("v4")
        keywords.add("install_only")
        keywords.add("tar.gz")
        keywords.add("cpython")
        keywords.add("x86_64")
        not_include.add("stripped")

    result = []
    for asset in assets:
        url = asset.url
        for k in not_include:
            if k in url:
                break
        else:
            for k in keywords:
                if k not in url:
                    break
            else:
                result.append(asset)
    if result:
        if len(result) > 1:
            # cpython-3.10.14
            def sort_k(i: Asset):
                # cpython-3.9.19
                m = re.search(r"-(\d+)\.(\d+)\.(\d+)", i.name)
                if m:
                    a, b, c = int(m[1]), int(m[2]), int(m[3])
                    return a * 10**3 + b * 10**2 + c * 10**1
                else:
                    return i.size

            result.sort(key=sort_k, reverse=True)
            for i in result:
                print(i.url)
            print(
                f"[{get_time()}] Got {len(result)} urls from github, will choose the first one and start download in 3 seconds. {result[0].name}"
            )
            for _ in range(3):
                print(3 - _, flush=True)
                time.sleep(1)
        return result[0]
    else:
        raise ValueError("No asset found.")


def download_python():
    """Download python portable interpreter from https://github.com/indygreg/python-build-standalone/releases. `python -m download_python -i` or `python -m download_python -a`(auto download the latest version matched the current platform: x86_64+install_only) or `python -m download_python -auto -k 3.11 -u`

    λ python -m download_python -i
    [10:56:17] Checking https://api.github.com/repos/indygreg/python-build-standalone/releases/latest
    [10:56:19] View the rules:
    https://gregoryszorc.com/docs/python-build-standalone/main/running.html#obtaining-distributions

    [10:56:19] Got 290 urls from github.

    [290] Enter keywords (can be int index or partial match, defaults to 0):
    0. windows
    1. linux
    2. darwin
    0
    [10:56:24] Filt with keyword: "windows". 290 => 40

    [40] Enter keywords (can be int index or partial match, defaults to 0):
    0. 3.12.3
    1. 3.11.9
    2. 3.10.14
    3. 3.9.19
    4. 3.8.19

    [10:56:25] Filt with keyword: "3.12.3". 40 => 8

    [8] Enter keywords (can be int index or partial match, defaults to 0):
    0. x86_64
    1. i686

    [10:56:28] Filt with keyword: "x86_64". 8 => 4

    [4] Enter keywords (can be int index or partial match, defaults to 0):
    0. shared-pgo-full.tar.zst
    1. shared-install_only.tar.gz
    2. pgo-full.tar.zst
    3. install_only.tar.gz
    3
    [10:56:33] Filt with keyword: "install_only.tar.gz". 4 => 1
    [10:56:33] Download URL: 39.1 MB
    https://github.com/indygreg/python-build-standalone/releases/download/20240415/cpython-3.12.3%2B20240415-x86_64-pc-windows-msvc-install_only.tar.gz
    File path to save(defaults to `./cpython-3.12.3+20240415-x86_64-pc-windows-msvc-install_only.tar.gz`)?
    or `q` to exit.

    [10:56:38] Start downloading...
    https://github.com/indygreg/python-build-standalone/releases/download/20240415/cpython-3.12.3%2B20240415-x86_64-pc-windows-msvc-install_only.tar.gz
    D:\github\morebuiltins\morebuiltins\download_python\cpython-3.12.3+20240415-x86_64-pc-windows-msvc-install_only.tar.gz
    [10:56:44] Downloading: 39.12 / 39.12 MB | 100.00% | 11.3 MB/s | 0s
    [10:56:44] Download complete."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--auto",
        "-a",
        help="auto choose the best choice(like platform), such as: -k 3.11 --auto",
        action="store_true",
    )
    parser.add_argument("--target", "-t", help="target dir path", default=".")
    parser.add_argument(
        "--keywords",
        "-k",
        help="keywords to filt, split by , for many keywords.",
        default="",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="interactive mode, will ask for filt keywords.",
    )
    parser.add_argument(
        "-u",
        "--unzip",
        action="store_true",
        help="unzip the tar.gz file to target/short_name/. only for .tar.gz",
    )
    args, _ = parser.parse_known_args()
    print(
        f"[{get_time()}] Checking https://api.github.com/repos/indygreg/python-build-standalone/releases/latest",
        flush=True,
    )
    assets = get_assets()
    print(
        f"[{get_time()}] View the rules:\nhttps://gregoryszorc.com/docs/python-build-standalone/main/running.html#obtaining-distributions\n",
        flush=True,
    )
    print(f"[{get_time()}] Got {len(assets)} urls from github.")
    keywords = args.keywords.split(",")
    if args.interactive:
        if args.keywords:
            assets = choose_asset(assets, keywords, auto=False)

        def sort_key(s):
            try:
                return tuple(map(int, s.split(".")))
            except ValueError:
                return s

        indexs = [4, 1, 5, 2, 3, 0, 6]
        for index in indexs:
            try:
                to_filt = sorted(
                    {i.keywords[index] for i in assets}, key=sort_key, reverse=True
                )
            except IndexError:
                continue
            if len(to_filt) == 1:
                continue
            choices = "\n".join((f"{idx}. {ii}" for idx, ii in enumerate(to_filt, 0)))
            arg = input(
                f"\n[{len(assets)}] Enter keywords (can be int index or partial match, defaults to 0):\n{choices}\n"
            )
            if not arg:
                arg = to_filt[0]
            elif arg.isdigit():
                arg = to_filt[int(arg)]
            old = len(assets)
            temp = [i for i in assets if arg == i.keywords[index]]
            if temp:
                assets = temp
            else:
                assets = [i for i in assets if arg in i.keywords[index]]
            print(
                f'[{get_time()}] Filt with keyword: "{arg}".',
                old,
                "=>",
                len(assets),
                flush=True,
            )
        while len(assets) > 1:
            names = "\n".join(i.name for i in assets)
            arg = input(f"Enter keyword to reduce the list (partial match):\n{names}\n")
            assets = [i for i in assets if arg in i.name]
        if not assets:
            input("No match, press enter to exit.")
            return
        asset = assets[0]
    else:
        asset = choose_asset(assets, keywords, args.auto)
    download_url = asset.url
    total_size = asset.size
    print(
        f"[{get_time()}] Download URL:",
        round(total_size / 1024**2, 1),
        "MB",
    )
    print(download_url, flush=True)
    target = Path(args.target).resolve()
    target.mkdir(parents=True, exist_ok=True)
    target_path = target / asset.name
    temp_path = target_path.with_name("python_downloading.tmp")
    for _path in [target_path, temp_path]:
        if _path.is_file():
            _path.unlink()
    print(f"[{get_time()}] Start downloading...")
    print(download_url)
    print(target_path.resolve(), flush=True)
    records = deque(maxlen=1000)
    last_print = time.time()

    def reporthook(blocknum, blocksize, totalsize):
        nonlocal last_print
        if totalsize < 0:
            totalsize = total_size
        _done = blocknum * blocksize
        if not _done:
            return
        percent = 100.0 * _done / totalsize
        total = totalsize / 1024 / 1024
        done = _done / 1024 / 1024 or total
        if percent > 100:
            percent = 100
        now = time.time()
        record = (now, _done)
        records.appendleft(record)
        timeleft = "-"
        _speed = 0
        if len(records) >= 2:
            for record in records:
                if now - record[0] > 1:
                    break
            time_diff = now - record[0]
            if time_diff:
                _speed = round((_done - record[1]) / time_diff, 1)
                secs = (totalsize - _done) / _speed
                if secs > 60:
                    timeleft = f"{int(secs / 60)}:{int(secs % 60):02}"
                else:
                    timeleft = f"{int(secs)}s"
        if _speed > 1024**2:
            speed = f"{round(_speed / 1024**2, 1)} MB/s"
        elif _speed > 1024:
            speed = f"{round(_speed / 1024, 1)} KB/s"
        else:
            speed = f"{round(_speed, 1)} B/s"
        if time.time() - last_print > 1:
            print(
                f"[{get_time()}] Downloading: {done:.2f} / {total:.2f} MB | {percent:.2f}% | {speed} | {timeleft} {' ' * 10}",
                end="\r",
                flush=True,
            )
            last_print = time.time()

    for _ in range(3):
        try:
            last_print = time.time()
            urllib.request.urlretrieve(
                download_url, temp_path.absolute().as_posix(), reporthook=reporthook
            )
            temp_path.rename(target_path)
            print(
                f"\n[{get_time()}] Download complete.",
                flush=True,
            )
            break
        except http.client.RemoteDisconnected:
            continue
        except KeyboardInterrupt:
            try:
                temp_path.unlink()
            except FileNotFoundError:
                pass
            print()
            print(f"\n[{get_time()}] Download canceled.", flush=True)
            return
        except Exception:
            print()
            traceback.print_exc()
            try:
                temp_path.unlink()
            except FileNotFoundError:
                pass
            break
    if args.interactive:
        print("Press enter to exit.", flush=True)
        return input()
    if args.unzip and target_path.name.endswith(".tar.gz") and target_path.is_file():
        target_dir = target_path.with_name(target_path.stem.split("+", 1)[0])
        if target_dir.is_dir():
            shutil.rmtree(target_dir.as_posix())
        print(get_time(), f"start unzip file to {target_dir.as_posix()}", flush=True)
        with tarfile.open(target_path.as_posix(), "r:gz") as tar:
            tar.extractall(path=target_dir.as_posix())
    print(get_time(), "bye~")


if __name__ == "__main__":
    download_python()



import json as _json
import re
import socket
import ssl
import threading
import time
from functools import lru_cache, partial
from http import HTTPStatus
from http.client import HTTPResponse
from pathlib import Path
from tempfile import gettempdir
from typing import Dict, Optional, Tuple, Union
from urllib.error import HTTPError, URLError
from urllib.parse import parse_qsl, quote_plus, urlencode, urlparse, urlunparse
from urllib.request import Request, urlopen, build_opener, ProxyHandler

__all__ = [
    "req",
    "DomainParser",
    "unparse_qsl",
    "update_url",
    "get_lan_ip",
    "make_response",
    "custom_dns",
]


def _request(
    url: str,
    params: Optional[dict] = None,
    headers: Optional[dict] = None,
    data: Optional[bytes] = None,
    json=None,
    form=None,
    timeout=None,
    method: str = "GET",
    verify=True,
    encoding=None,
    proxy=None,
    urlopen_kwargs: Optional[dict] = None,
    build_opener_handlers: Optional[list] = None,
    **kwargs,
):
    if params:
        if "?" in url:
            sep = "&"
        else:
            sep = "?"
        url += f"{sep}{urlencode(params)}"
    if headers is None:
        headers = {}
    if data:
        if isinstance(data, bytes):
            headers.setdefault("Content-Type", "")
        elif isinstance(data, dict):
            # as form
            data = urlencode(data, doseq=True).encode("utf-8")
            headers.setdefault("Content-Type", "application/x-www-form-urlencoded")
    else:
        if json:
            data = _json.dumps(json, allow_nan=False).encode("utf-8")
            headers.setdefault("Content-Type", "application/json")
        elif form:
            data = urlencode(form, doseq=True).encode("utf-8")
            headers.setdefault("Content-Type", "application/x-www-form-urlencoded")
    req = Request(method=method, url=url, headers=headers, data=data, **kwargs)
    if urlopen_kwargs is None:
        urlopen_kwargs = {}
    urlopen_kwargs["url"] = req
    if timeout is not None:
        urlopen_kwargs.setdefault("timeout", timeout)
    if not verify:
        ctx = urlopen_kwargs.setdefault("context", ssl.create_default_context())
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

    def ensure_attrs():
        if hasattr(HTTPResponse, "get_encoding"):
            return True

        def get_body(self):
            return getattr(self, "_response_body", b"")

        def get_encoding(self):
            charset = re.search(rb"charset[\s\"'=]*([a-zA-Z0-9._-]+)", self.content)
            if charset:
                encoding = charset.group(1).decode("utf-8", "ignore")
            else:
                encoding = "utf-8"
            return encoding

        def get_text(self):
            if self.encoding is None:
                encoding = self.get_encoding()
            else:
                encoding = self.encoding
            return self.content.decode(encoding, "replace")

        def get_json(self, **kwargs):
            return _json.loads(self.content, **kwargs)

        def get_code(self):
            return self.code

        def get_ok(self):
            return self.code < 400

        for cls in [HTTPResponse, HTTPError]:
            # monkey patch
            setattr(cls, "encoding", None)
            setattr(cls, "status_code", property(get_code))
            setattr(cls, "content", property(get_body))
            setattr(cls, "text", property(get_text))
            setattr(cls, "ok", property(get_ok))
            setattr(cls, "json", get_json)
            setattr(cls, "get_encoding", get_encoding)

    ensure_attrs()
    response = None
    try:
        build_opener_handlers = build_opener_handlers or []
        if proxy:
            if build_opener_handlers:
                raise ValueError(
                    "proxy and build_opener_handlers are mutually exclusive"
                )
            proxy_support = ProxyHandler({"http": proxy, "https": proxy})
            build_opener_handlers.append(proxy_support)
        if build_opener_handlers:
            opener = build_opener(*build_opener_handlers)
            resp = opener.open(req, data, timeout)
        else:
            resp = urlopen(**urlopen_kwargs)
        with resp as resp:
            body = resp.read()
            response = resp
    except HTTPError as error:
        body = error.read()
        response = error
    setattr(response, "_response_body", body)
    setattr(response, "encoding", encoding)
    setattr(response, "headers", dict(response.headers))
    return response


class req:
    """A basic mock for requests, performant albeit simplistic.

    >>> import time
    >>> r = req.get("https://postman-echo.com/get?a=2", timeout=3, params={"b": "3"})
    >>> r.url
    'https://postman-echo.com/get?a=2&b=3'
    >>> r.ok
    True
    >>> r.status_code
    200
    >>> r.text.startswith('{')
    True
    >>> r = req.post("https://postman-echo.com/post?a=2", timeout=3, params={"b": "3"}, data=b"mock data")
    >>> r.json()["data"]
    'mock data'
    >>> r.json()["args"]
    {'a': '2', 'b': '3'}
    >>> r = req.post("https://postman-echo.com/post?a=2", timeout=3, json={"data": "yes json"})
    >>> r.json()["json"]
    {'data': 'yes json'}
    """

    RequestErrors = (URLError,)

    request = _request
    get = partial(_request, method="GET")
    head = partial(_request, method="HEAD")
    post = partial(_request, method="POST")
    put = partial(_request, method="PUT")
    delete = partial(_request, method="DELETE")
    patch = partial(_request, method="PATCH")
    options = partial(_request, method="OPTIONS")


class DomainParser(object):
    """Extracts the Second-level domain (SLD) from a provided hostname or URL.

    >>> domain_parser = DomainParser()
    >>> domain_parser.parse_hostname("github.com")
    'github.com'
    >>> domain_parser.parse_hostname("www.github.com")
    'github.com'
    >>> domain_parser.parse_hostname("www.api.github.com.cn")
    'github.com.cn'
    >>> domain_parser.parse_hostname("a.b.c.kawasaki.jp")
    'c.kawasaki.jp'
    >>> domain_parser.parse_hostname("a.b.c.city.kawasaki.jp")
    'c.city.kawasaki.jp'
    >>> domain_parser.parse_hostname("a.bbbbbb.cccccc")
    ''
    >>> domain_parser.parse_hostname("a.bbbbbb.cccccc", default="b.c")
    'b.c'
    >>> domain_parser.parse_url("https://github.com/ClericPy/morebuiltins")
    'github.com'

    """

    # default cache path, avoid request too many times.
    # you can reset this variable with new path or use $TMP path by default.
    PUBLIC_SUFFIX_CACHE_PATH = Path(gettempdir()) / "public_suffix_list.dat"
    PUBLIC_SUFFIX_API_1 = "https://publicsuffix.org/list/public_suffix_list.dat"
    PUBLIC_SUFFIX_API_2 = (
        "https://github.com/publicsuffix/list/raw/master/public_suffix_list.dat"
    )

    def __init__(self, cache_size=0, public_suffix_file_path=...):
        if public_suffix_file_path is ...:
            public_suffix_file_path = self.PUBLIC_SUFFIX_CACHE_PATH
        self.public_suffix_file_path = Path(public_suffix_file_path)
        self.init_local_cache()
        # use lru_cache for performance
        if cache_size:
            self._parse_cached = lru_cache(maxsize=cache_size)(self._parse)
        else:
            self._parse_cached = self._parse

    def parse_url(self, url: str, default=""):
        return self.parse_hostname(urlparse(url).netloc, default=default)

    def parse_hostname(self, hostname: str, default=""):
        return self._parse_cached(hostname=hostname, default=default)

    def _parse(self, hostname: str, default=""):
        parts = hostname.split(".")
        _suffix_trie = self._suffix_trie
        finish = False
        offset = 0
        for index in range(len(parts) - 1, -1, -1):
            part = parts[index]
            try:
                offset = index
                if finish:
                    break
                _suffix_trie = _suffix_trie[part]
            except KeyError:
                if "*" in _suffix_trie:
                    finish = True
                break
        result = parts[offset:]
        if len(result) > 1:
            return ".".join(result)
        else:
            return default

    def init_local_cache(self):
        if not self.public_suffix_file_path.is_file():
            self.public_suffix_file_path.parent.mkdir(parents=True, exist_ok=True)
            headers = {"User-Agent": ""}
            for api in [self.PUBLIC_SUFFIX_API_1, self.PUBLIC_SUFFIX_API_2]:
                try:
                    req = Request(method="GET", url=api, headers=headers)
                    with urlopen(req, timeout=10) as r:
                        text = r.read().decode("utf-8")
                        self.public_suffix_file_path.write_text(text, encoding="utf-8")
                        break
                except URLError:
                    continue
            else:
                raise ValueError("request from public suffix api failed")
        _trie = {}
        with open(self.public_suffix_file_path, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith("/"):
                    parts = line.split(".")
                    root = _trie
                    for index in range(len(parts) - 1, -1, -1):
                        part = parts[index]
                        # disable * while exist !
                        if part[0] == "!":
                            root = root.setdefault(part[1:], {})
                            break
                        else:
                            root = root.setdefault(part, {})
        self._suffix_trie = _trie


def unparse_qsl(qsl, sort=False, reverse=False):
    """Provides the inverse operation of parse_qsl, converting query string lists back into a URL-encoded string."""
    result = []
    if sort:
        qsl = sorted(qsl, key=lambda x: x[0], reverse=reverse)
    for key, value in qsl:
        query_name = quote_plus(key)
        result.append(query_name + "=" + quote_plus(value))
    return "&".join(result)


def update_url(
    url,
    params: Optional[dict] = None,
    replace_kwargs=None,
    sort=False,
    reverse=False,
    **kw_params,
):
    """Organizes the query arguments within a URL to standardize its format.

    >>> update_url('http://www.github.com?b=1&c=1&a=1', {"b": None, "c": None})  # remove params
    'http://www.github.com?a=1'
    >>> update_url("http://www.github.com?b=1&c=1&a=1", a="123", b=None)  # update params with kwargs
    'http://www.github.com?c=1&a=123'
    >>> update_url('http://www.github.com?b=1&c=1&a=1', sort=True)  # sort params
    'http://www.github.com?a=1&b=1&c=1'
    >>> update_url("http://www.github.com?b=1&c=1&a=1", {"a": "999"})  # update params
    'http://www.github.com?b=1&c=1&a=999'
    >>> update_url("http://www.github.com?b=1&c=1&a=1", replace_kwargs={"netloc": "www.new_host.com"})  # update netloc
    'http://www.new_host.com?b=1&c=1&a=1'

    replace_kwargs is a dict to update attributes before sorting  (such as scheme / netloc...).
    """
    parsed = urlparse(url)
    if replace_kwargs is None:
        replace_kwargs = {}
    todo = parse_qsl(parsed.query)
    if params:
        kw_params.update(params)
    if kw_params:
        delete_indexes = set()
        for index, item in enumerate(todo):
            name, _ = item
            if name in kw_params:
                new = kw_params.pop(name)
                if new is None:
                    delete_indexes.add(index)
                else:
                    todo[index] = (name, new)
        if delete_indexes:
            todo = [
                item for index, item in enumerate(todo) if index not in delete_indexes
            ]
        for k, v in kw_params.items():
            todo.append((k, v))
    sorted_parsed = parsed._replace(
        query=unparse_qsl(todo, sort=sort, reverse=reverse), **replace_kwargs
    )
    return urlunparse(sorted_parsed)


def get_lan_ip():
    try:
        # 获取本机主机名
        hostname = socket.gethostname()
        # 获取本机IP
        return socket.gethostbyname(hostname)
    except Exception:
        pass
    return ""


def make_response(
    body: Union[str, bytes, list, dict] = "",
    status: int = 200,
    protocol="HTTP/1.1",
    headers=None,
    encoding="utf-8",
) -> bytes:
    """Generates an HTTP response based on the provided parameters.

    :param body: The response body which can be a string, bytes, list, or dictionary. Default is an empty string.
    :param status: The HTTP status code. Default is 200.
    :param protocol: The HTTP protocol version. Default is "HTTP/1.1".
    :param headers: A dictionary of HTTP response headers. Default is None.
    :param encoding: The encoding to use. Default is "utf-8".
    :return: A byte sequence representing the constructed HTTP response.
    """
    # b'HTTP/1.1 200 OK\r\nDate: Fri, 21 Oct 2022 01:04:34 GMT\r\nContent-Length: 2\r\nContent-Type: text/plain; charset=utf-8\r\n\r\nok'
    if headers:
        headers_string = ""
        for k, v in headers.items():
            headers_string += f"\r\n{k.title()}: {v}"
    else:
        headers_string = ""
    struct_time = time.gmtime(time.time())
    now = time.strftime("%a, %d %b %Y %H:%M:%S GMT", struct_time)

    if isinstance(body, bytes):
        _body = body
    elif isinstance(body, str):
        _body = body.encode(encoding, errors="replace")
    else:
        _body = _json.dumps(body, default=str, ensure_ascii=False).encode(
            encoding, errors="replace"
        )
    http_status = HTTPStatus(status)
    head = f"{protocol} {http_status.numerator} {http_status.name}\r\nDate: {now}\r\nContent-Length: {len(_body)}{headers_string}\r\n\r\n".encode(
        encoding, errors="replace"
    )
    return head + _body


original_getaddrinfo = socket.getaddrinfo


def custom_dns(custom: Dict[Union[str, Tuple[str, int]], Tuple[str, int]], thread=True):
    """Custom the DNS of socket.getaddrinfo, only effect current thread.

    [WARNING] This will modify the global socket.getaddrinfo.

    >>> from concurrent.futures import ThreadPoolExecutor
    >>> # this only effect current thread
    >>> custom_dns({"1.1.1.1": ("127.0.0.1", 80), ("1.1.1.1", 80): ("192.168.0.1", 443)})
    >>> socket.getaddrinfo('1.1.1.1', 80)[0][-1]
    ('192.168.0.1', 443)
    >>> socket.getaddrinfo('1.1.1.1', 8888)[0][-1]
    ('127.0.0.1', 80)
    >>> ThreadPoolExecutor().submit(lambda : socket.getaddrinfo('1.1.1.1', 8888)[0][-1]).result()
    ('1.1.1.1', 8888)
    >>> # this effect global socket.getaddrinfo
    >>> custom_dns({"1.1.1.1": ("127.0.0.1", 80), ("1.1.1.1", 80): ("192.168.0.1", 443)}, thread=False)
    >>> ThreadPoolExecutor().submit(lambda : socket.getaddrinfo('1.1.1.1', 8888)[0][-1]).result()
    ('127.0.0.1', 80)

    Demo:

        custom_dns(custom={("MY_PROXY_HOST", 80): ("xxxxxxxxx", 43532)})
        print(
            requests.get(
                "https://www.github.com/", proxies={"all": "http://MY_PROXY_HOST"}
            ).text
        )
    """
    if thread:
        thread = threading.current_thread()
    else:
        thread = None

    def wrapper(host, port=80, family=0, type=0, proto=0, flags=0):
        if not thread or thread is threading.current_thread():
            host, port = custom.get((host, port)) or custom.get(host) or (host, port)
        return original_getaddrinfo(host, port, family, type, proto, flags)

    socket.getaddrinfo = wrapper


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os
import re
import subprocess
import sys
import tempfile
import time
import typing
import venv
from pathlib import Path


def ttime(
    timestamp=None, tzone=int(-time.timezone / 3600), fail="", fmt="%Y-%m-%d %H:%M:%S"
):
    fix_tz = tzone * 3600
    if timestamp is None:
        timestamp = time.time()
    else:
        timestamp = float(timestamp)
        if 1e12 <= timestamp < 1e13:
            # Compatible timestamp with 13-digit milliseconds
            timestamp = timestamp / 1000
    try:
        timestamp = time.time() if timestamp is None else timestamp
        return time.strftime(fmt, time.gmtime(timestamp + fix_tz))
    except Exception:
        return fail


class FreezeTool(object):
    VENV_NAME = "zipapps_venv"
    # not stable
    FASTER_PREPARE_PIP = False

    def __init__(self, output: str, pip_args: list):
        if not pip_args:
            raise RuntimeError("pip args is null")
        self.temp_dir: typing.Optional[tempfile.TemporaryDirectory] = None
        self.pip_args = pip_args
        self.output_path = output

    def log(self, msg, flush=False):
        _msg = f"{ttime()} | {msg}"
        print(_msg, file=sys.stderr, flush=flush)

    def run(self):
        self.log(
            "All the logs will be redirected to stderr to ensure the output is stdout."
        )
        self.temp_dir = tempfile.TemporaryDirectory(prefix="zipapps_")
        self.temp_path = Path(self.temp_dir.name)
        self.log(
            f"Start mkdir temp folder: {self.temp_path.absolute()}, exist={self.temp_path.is_dir()}"
        )
        self.install_env()
        output = self.install_packages()
        self.freeze_requirements(output)
        return output

    def install_env(self):
        venv_path = self.temp_path / self.VENV_NAME
        self.log(f"Initial venv with pip: {venv_path.absolute()}")
        if self.FASTER_PREPARE_PIP:
            venv.create(env_dir=venv_path, system_site_packages=False, with_pip=False)
            import shutil

            import pip

            pip_dir = Path(pip.__file__).parent
            if os.name == "nt":
                target = venv_path / "Lib" / "site-packages" / "pip"
            else:
                pyv = "python%d.%d" % sys.version_info[:2]
                target = venv_path / "lib" / pyv / "site-packages" / "pip"
            shutil.copytree(pip_dir, target)
        else:
            venv.create(env_dir=venv_path, system_site_packages=False, with_pip=True)
        if not venv_path.is_dir():
            raise FileNotFoundError(str(venv_path))

    def install_packages(self):
        if os.name == "nt":
            python_path = self.temp_path / self.VENV_NAME / "Scripts" / "python.exe"
        else:
            python_path = self.temp_path / self.VENV_NAME / "bin" / "python"
        args = [
            str(python_path.absolute()),
            "-m",
            "pip",
            "install",
        ] + self.pip_args
        self.log(f'Install packages in venv: {args}\n{"-" * 30}')
        with subprocess.Popen(args, stdout=subprocess.PIPE) as proc:
            for line in proc.stdout:
                try:
                    line = line.decode()
                except ValueError:
                    line = line.decode("utf-8", "ignore")
                print(line.rstrip(), file=sys.stderr, flush=True)
        args = [str(python_path.absolute()), "-m", "pip", "freeze"]
        print("-" * 30, file=sys.stderr)
        self.log(f"Freeze packages in venv: {args}")
        output = subprocess.check_output(args)
        try:
            result = output.decode("utf-8")
        except ValueError:
            result = output.decode()
        result = re.sub("(\n|\r)+", "\n", result).strip()
        return result

    def freeze_requirements(self, output):
        if self.output_path == "-":
            print(output, flush=True)
        else:
            print(output, file=sys.stderr, flush=True)
            with open(self.output_path, "w", encoding="utf-8") as f:
                print(output, file=f, flush=True)

    def remove_env(self):
        if self.temp_dir and self.temp_path.is_dir():
            self.temp_dir.cleanup()
            self.log(
                f"Delete temp folder: {self.temp_path.absolute()}, exist={self.temp_path.is_dir()}"
            )

    def __del__(self):
        self.remove_env()

    def __enter__(self):
        return self

    def __exit__(self, *e):
        self.remove_env()


def test():
    with FreezeTool("-", ["six==1.15.0"]) as ft:
        result = ft.run()
        # print(result)
        assert result == "six==1.15.0"


if __name__ == "__main__":
    test()



import asyncio
import json
import logging
import logging.handlers
import os
import signal
import sys
import time
import traceback
import typing
from collections import Counter, namedtuple
from concurrent.futures import Future
from pathlib import Path
from queue import Empty, Queue
from threading import Thread, Timer

from ..functools import RotatingFileWriter
from ..ipc import SocketLogHandlerEncoder, SocketServer
from ..utils import format_error, read_size, ttime

__all__ = ["LogServer"]

QueueMsg = namedtuple("QueueMsg", ["name", "text", "file_args"])


class LogServer(SocketServer):
    """Log Server for SocketHandler, create a socket server with asyncio.start_server. Update settings of rotation/formatter with extra: {"max_size": 1024**2, "formatter": logging.Formatter(fmt="%(asctime)s - %(filename)s - %(message)s")}

    Server side:
        python -m morebuiltins.cmd.log_server --log-dir=./logs --host 127.0.0.1 --port 8901

    Client side:

    ```python
    import logging
    import logging.handlers

    logger = logging.getLogger("client")
    logger.setLevel(logging.DEBUG)
    h = logging.handlers.SocketHandler("127.0.0.1", 8901)
    h.setLevel(logging.DEBUG)
    logger.addHandler(h)
    for _ in range(5):
        logger.info(
            "hello world!",
            extra={
                "max_size": 1024**2,
                "formatter": logging.Formatter(
                    fmt="%(asctime)s - %(filename)s - %(message)s"
                ),
            },
        )
    # [client] 2024-08-10 19:30:07,113 - temp3.py - hello world!
    # [client] 2024-08-10 19:30:07,113 - temp3.py - hello world!
    # [client] 2024-08-10 19:30:07,113 - temp3.py - hello world!
    # [client] 2024-08-10 19:30:07,113 - temp3.py - hello world!
    # [client] 2024-08-10 19:30:07,114 - temp3.py - hello world!
    ```

    More docs:
        python -m morebuiltins.cmd.log_server -h
        usage: log_server.py [-h] [--host HOST] [--port PORT] [--log-dir LOG_DIR] [--name NAME] [--server-log-args SERVER_LOG_ARGS] [--handle-signals HANDLE_SIGNALS] [--max-queue-size MAX_QUEUE_SIZE]
                            [--max-queue-buffer MAX_QUEUE_BUFFER] [--log-stream LOG_STREAM]

        options:
        -h, --help            show this help message and exit
        --host HOST
        --port PORT
        --log-dir LOG_DIR     log dir to save log files, if empty, log to stdout with --log-stream
        --name NAME           log server name
        --server-log-args SERVER_LOG_ARGS
                                max_size,max_backups for log files, default: 10485760,5 == 10MB each log file, 1 name.log + 5 backups
        --handle-signals HANDLE_SIGNALS
        --max-queue-size MAX_QUEUE_SIZE
                                max queue size for log queue, log will be in memory queue before write to file
        --max-queue-buffer MAX_QUEUE_BUFFER
                                chunk size of lines before write to file
        --log-stream LOG_STREAM
                                log to stream, if --log-stream='' will mute the stream log

    """

    def __init__(
        self,
        host="127.0.0.1",
        port=8901,
        log_dir=None,
        name="log_server",
        server_log_args=(10 * 1024**2, 5),
        handle_signals=(2, 15),
        max_queue_size=100000,
        max_queue_buffer=1000,
        log_stream=sys.stdout,
    ):
        super().__init__(
            host,
            port,
            handler=self.default_handler,
            encoder=SocketLogHandlerEncoder(),
            start_callback=self.start_callback,
            end_callback=self.end_callback,
        )
        self.name = name
        self.server_log_args = {
            "max_size": server_log_args[0],
            "max_backups": server_log_args[1],
        }
        self.log_stream = log_stream
        self.log_dir = Path(log_dir).resolve() if log_dir else None
        if self.log_dir:
            self.log_dir.mkdir(exist_ok=True, parents=True)
        self.opened_files: typing.Dict[str, RotatingFileWriter] = {}
        self.handle_signals = handle_signals
        for sig in handle_signals:
            signal.signal(sig, self.handle_signal)
        self.max_queue_size = max_queue_size
        self.write_queue = Queue(maxsize=max_queue_size)
        self.max_queue_buffer = max_queue_buffer
        self.file_writer_thread = Thread(target=self.write_worker)

        self._shutdown_signals = 0
        self._thread_done = Future()

    def end_callback(self):
        self.send_log("stopping log server")
        self.write_queue.put(None)
        try:
            self._thread_done.result(timeout=5)
        except TimeoutError:
            os._exit(1)

    def start_callback(self):
        self.send_log(
            f"started log server on {self.host}:{self.port}, handle_signals={self.handle_signals}, max_queue_size={self.max_queue_size}, max_queue_buffer={self.max_queue_buffer}, log_stream={getattr(self.log_stream, 'name', None)}, log_dir={self.log_dir}"
        )
        self.file_writer_thread.start()

    def send_log(
        self, msg: str, error: typing.Optional[Exception] = None, level="INFO"
    ):
        now = time.time()
        ms = str(now).split(".", 1)[1][:3]
        if error:
            msg = f"{msg} | {format_error(error)}"
        msg = f"{ttime(now)},{ms} | {level: >5} | {msg}"
        self.write_queue.put((self.name, msg, self.server_log_args))

    def get_targets(self, name: str, max_size=5 * 1024**2, max_backups=1):
        targets = []
        if self.log_stream:
            targets.append(self.log_stream)
        if self.log_dir:
            if name in self.opened_files:
                fw = self.opened_files[name]
                if fw.max_size != max_size:
                    fw.max_size = max_size
                if fw.max_backups != max_backups:
                    fw.max_backups = max_backups
                targets.append(fw)
            else:
                try:
                    fw = RotatingFileWriter(
                        self.log_dir.joinpath(f"{name}.log"),
                        max_size=max_size,
                        max_backups=max_backups,
                    )
                except Exception as e:
                    self.send_log(
                        f"error in get_targets({name!r}, {max_size!r}, {max_backups!r})",
                        e,
                        level="ERROR",
                    )
                targets.append(
                    self.opened_files.setdefault(
                        name,
                        fw,
                    )
                )
        return targets

    def write_worker(self):
        last_log_time = time.time()
        interval = 60
        lines_counter = Counter()
        size_counter = Counter()
        shutdown = False
        self.send_log("started write_worker daemon")
        while not shutdown:
            try:
                new_lines = {}
                timeout = 1
                for _ in range(self.max_queue_buffer):
                    try:
                        q_msg: QueueMsg = self.write_queue.get(timeout=timeout)
                        timeout = 0
                        if q_msg is None:
                            if not shutdown:
                                shutdown = True
                                self.send_log("stopping write_worker daemon")
                            continue
                        name, line, file_args = q_msg
                        if name in new_lines:
                            data = new_lines[name]
                            data["file_args"] = file_args
                        else:
                            data = {"file_args": file_args, "lines": []}
                            new_lines[name] = data
                        data["lines"].append(line)
                    except Empty:
                        break
                for name, data in new_lines.items():
                    file_args = data["file_args"]
                    targets = self.get_targets(name, **file_args)
                    for log_file in targets:
                        if log_file is self.log_stream:
                            head = f"[{name}] "
                        else:
                            head = ""
                        try:
                            for line in data["lines"]:
                                text = f"{head}{line}\n"
                                log_file.write(text)
                            # only flush once per file
                            log_file.flush()
                        except Exception as e:
                            self.send_log(
                                f"error in write_worker ({name})", e, level="WARN"
                            )
                    if name != self.name:
                        lines_counter[name] += len(data["lines"])
                        size_counter[name] += sum(
                            [len(line) for line in data["lines"]]
                        ) + len(data["lines"])
                if lines_counter:
                    now = time.time()
                    if now - last_log_time > interval:
                        start = ttime(last_log_time, fmt="%H:%M:%S")
                        end = ttime(now, fmt="%H:%M:%S")
                        lines_msg = json.dumps(
                            {
                                name: value
                                for name, value in lines_counter.most_common(10)
                            },
                            ensure_ascii=False,
                        )
                        size_msg = json.dumps(
                            {
                                name: read_size(value, 1, shorten=True)
                                for name, value in size_counter.most_common(10)
                            },
                            ensure_ascii=False,
                        )
                        self.send_log(
                            f"{start} - {end} log counter: {sum(lines_counter.values())} lines ({lines_msg}), {read_size(sum(size_counter.values()), 1, shorten=True)} ({size_msg})"
                        )
                        lines_counter.clear()
                        last_log_time = now
            except Exception as e:
                self.send_log("error in write_worker", e, level="ERROR")
                print(format_error(e), file=sys.stderr, flush=True)
                traceback.print_exc()
                self.shutdown()
                break
        self._thread_done.set_result(None)

    async def default_handler(self, record: dict):
        # record demo:
        # {"name": "test_logger", "msg": "test socket logging message", "args": null, "levelname": "INFO", "levelno": 20, "pathname": "/PATH/temp.py", "filename": "temp.py", "module": "temp", "exc_info": null, "exc_text": null, "stack_info": null, "lineno": 38, "funcName": "main", "created": 1723270162.5119407, "msecs": 511.0, "relativeCreated": 102.74338722229004, "thread": 8712, "threadName": "MainThread", "processName": "MainProcess", "process": 19104}
        try:
            if not isinstance(record, dict):
                raise TypeError("item must be a dict")
            name = record["name"]
            if "formatter" in record:
                text = record["formatter"].format(logging.LogRecord(level=0, **record))
            else:
                text = record["msg"]
            file_args = {
                k: v for k, v in record.items() if k in {"max_size", "max_backups"}
            }
            q_msg = QueueMsg(name=name, text=text, file_args=file_args)
            self.write_queue.put_nowait(q_msg)
        except Exception as e:
            self.send_log("error in default_handler", e, level="WARN")
        finally:
            del record

    def handle_signal(self, sig, frame):
        self._shutdown_signals += 1
        if self._shutdown_signals > 5:
            os._exit(1)
        msg = f"received signal: {sig}, count: {self._shutdown_signals}"
        self.send_log(msg)
        self.shutdown()
        self.write_queue.put(None)

    def __del__(self):
        for f in self.opened_files.values():
            f.close()


async def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="127.0.0.1")
    parser.add_argument("--port", default=8901, type=int)
    parser.add_argument(
        "--log-dir",
        default="",
        dest="log_dir",
        help="log dir to save log files, if empty, log to stdout with --log-stream",
    )
    parser.add_argument("--name", default="log_server", help="log server name")
    parser.add_argument(
        "--server-log-args",
        default="10485760,5",
        dest="server_log_args",
        help="max_size,max_backups for log files, default: 10485760,5 == 10MB each log file, 1 name.log + 5 backups",
    )
    parser.add_argument("--handle-signals", default="2,15", dest="handle_signals")
    parser.add_argument(
        "--max-queue-size",
        default=100000,
        type=int,
        help="max queue size for log queue, log will be in memory queue before write to file",
    )
    parser.add_argument(
        "--max-queue-buffer",
        default=1000,
        type=int,
        help="chunk size of lines before write to file",
    )
    parser.add_argument(
        "--log-stream",
        default="sys.stdout",
        help="log to stream, if --log-stream='' will mute the stream log",
    )
    args = parser.parse_args()
    log_stream = {"sys.stdout": sys.stdout, "sys.stderr": sys.stderr, "": ""}[
        args.log_stream
    ]
    async with LogServer(
        host=args.host,
        port=args.port,
        log_dir=args.log_dir,
        name=args.name,
        server_log_args=tuple(map(int, args.server_log_args.split(","))),
        handle_signals=tuple(map(int, args.handle_signals.split(","))),
        max_queue_size=args.max_queue_size,
        max_queue_buffer=args.max_queue_buffer,
        log_stream=log_stream,
    ) as ls:
        await ls.wait_closed()


if __name__ == "__main__":
    asyncio.run(main())
